"""Abstract class for IR Passes for Jac."""
from typing import Optional, TypeVar

import jaclang.jac.absyntree as ast
from jaclang.jac.transform import Transform
from jaclang.utils.helpers import pascal_to_snake

T = TypeVar("T", bound=ast.AstNode)


class Pass(Transform):
    """Abstract class for IR passes."""

    def __init__(
        self,
        mod_path: str,
        input_ir: ast.AstNode,
        base_path: str = "",
        prior: Optional[Transform] = None,
    ) -> None:
        """Initialize parser."""
        self.term_signal = False
        self.prune_signal = False
        self.cur_node = input_ir  # tracks current node during traversal
        Transform.__init__(self, mod_path, input_ir, base_path, prior)

    def before_pass(self) -> None:
        """Run once before pass."""
        pass

    def after_pass(self) -> None:
        """Run once after pass."""
        pass

    def enter_node(self, node: ast.AstNode) -> None:
        """Run on entering node."""
        if hasattr(self, f"enter_{pascal_to_snake(type(node).__name__)}"):
            getattr(self, f"enter_{pascal_to_snake(type(node).__name__)}")(node)
        if isinstance(node, ast.Parse) and hasattr(self, f"enter_{node.name}"):
            getattr(self, f"enter_{node.name}")(node)

    def exit_node(self, node: ast.AstNode) -> None:
        """Run on exiting node."""
        if hasattr(self, f"exit_{pascal_to_snake(type(node).__name__)}"):
            getattr(self, f"exit_{pascal_to_snake(type(node).__name__)}")(node)
        if isinstance(node, ast.Parse) and hasattr(self, f"exit_{node.name}"):
            getattr(self, f"exit_{node.name}")(node)

    def terminate(self) -> None:
        """Terminate traversal."""
        self.term_signal = True

    def prune(self) -> None:
        """Prune traversal."""
        self.prune_signal = True

    def get_all_sub_nodes(
        self, node: ast.AstNode, typ: type[T], brute_force: bool = False
    ) -> list[T]:
        """Get all sub nodes of type."""
        result = []
        # Assumes pass built the sub node table
        if not node:
            return result
        elif len(node._sub_node_tab) and not brute_force:
            result.extend(node._sub_node_tab[typ] if typ in node._sub_node_tab else [])
        elif len(node.kid):
            if not brute_force:
                raise ValueError(f"Node has no sub_node_tab. {node}")
            # Brute force search
            else:
                for i in node.kid:
                    if isinstance(i, typ):
                        result.append(i)
                    result.extend(self.get_all_sub_nodes(i, typ, brute_force))
        return result

    def recalculate_parents(self, node: ast.AstNode) -> None:
        """Recalculate parents."""
        if not node:
            return
        for i in node.kid:
            if i:
                i.parent = node
                self.recalculate_parents(i)

    # Transform Implementations
    # -------------------------
    def transform(self, ir: ast.AstNode) -> ast.AstNode:
        """Run pass."""
        # Only performs passes on proper ASTs
        if not isinstance(ir, ast.AstNode):
            return ir
        self.before_pass()
        if not isinstance(ir, ast.AstNode):
            raise ValueError("Current node is not an AstNode.")
        self.traverse(ir)
        # Checks if self.ir is created during traversal
        self.ir = self.ir if hasattr(self, "ir") else ir
        self.after_pass()
        return self.ir

    def traverse(self, node: ast.AstNode) -> ast.AstNode:
        """Traverse tree."""
        if self.term_signal:
            return node
        self.cur_node = node
        self.enter_node(node)
        if not self.prune_signal:
            for i in node.kid:
                if i:
                    self.traverse(i)
        else:
            self.prune_signal = False
        self.cur_node = node
        self.exit_node(node)
        return node

    def update_code_loc(self, node: Optional[ast.AstNode] = None) -> None:
        """Update code location."""
        if node is None:
            node = self.cur_node
        if not isinstance(node, ast.AstNode):
            self.ice("Current node is not an AstNode.")
        self.cur_line = node.line
        if node.mod_link:
            self.rel_mod_path = node.mod_link.rel_mod_path

    def error(self, msg: str) -> None:
        """Pass Error."""
        self.update_code_loc()
        self.log_error(f"{msg}")

    def warning(self, msg: str) -> None:
        """Pass Error."""
        self.update_code_loc()
        self.log_warning(f"{msg}")

    def ice(self, msg: str) -> None:
        """Pass Error."""
        if isinstance(self.cur_node, ast.AstNode):
            self.cur_line = self.cur_node.line
        self.log_error(f"ICE: Pass {self.__class__.__name__} - {msg}")
        raise RuntimeError(
            f"Internal Compiler Error: Pass {self.__class__.__name__} - {msg}"
        )


class PrinterPass(Pass):
    """Printer Pass for Jac AST."""

    def enter_node(self, node: ast.AstNode) -> None:
        """Run on entering node."""
        print("Entering:", node)
        super().enter_node(node)

    def exit_node(self, node: ast.AstNode) -> None:
        """Run on exiting node."""
        super().exit_node(node)
        print("Exiting:", node)



"""Tests for Jac lexer."""
from typing import Generator

from jaclang.jac.constant import Tokens
from jaclang.jac.lexer import JacLexer
from jaclang.utils.test import TestCase


class TestLexer(TestCase):
    """Test Jac lexer."""

    def test_lexer(self) -> None:
        """Basic test for lexer."""
        tokens = []
        ir = JacLexer(mod_path="", input_ir=self.load_fixture("lexer_fam.jac")).ir
        if not isinstance(ir, Generator):
            raise ValueError("Lexer did not return generator.")
        for t in ir:
            tokens.append(t)
        self.assertEqual(tokens[0].type, "DOC_STRING")
        self.assertEqual(
            [
                tokens[10].type,
                tokens[10].value,
                tokens[10].lineno,
                tokens[10].index,
                tokens[10].end,
            ],
            ["KW_FROM", "from", 9, 140, 144],
        )
        self.assertEqual(
            [
                tokens[-1].type,
                tokens[-1].value,
                tokens[-1].lineno,
            ],
            ["RBRACE", "}", 61],
        )

    def test_col_idxs(self) -> None:
        """Basic test for lexer."""
        tokens = []
        ir = JacLexer(mod_path="", input_ir=self.load_fixture("lexer_fam.jac")).ir
        if not isinstance(ir, Generator):
            raise ValueError("Lexer did not return generator.")
        for t in ir:
            tokens.append((t.value, t.lineno, t.index - t.lineidx, t.end - t.lineidx))
        self.assertEqual(tokens[12], ("activity", 9, 16, 24))
        self.assertEqual(tokens[-3], ("outside_func", 59, 24, 36))

    def test_enum_matches_lexer_toks(self) -> None:
        """Test that enum stays synced with lexer."""
        for token in JacLexer.tokens:
            self.assertIn(token, Tokens.__members__)
        for token in Tokens:
            self.assertIn(token.name, JacLexer.tokens)
        for token in Tokens:
            self.assertIn(token.value, JacLexer.tokens)



"""Jac Symbol Table."""
import pprint
from typing import Optional, TypeVar


import jaclang.jac.absyntree as ast


class Symbol:
    """Symbol."""

    def __init__(
        self,
        name: str,
        node: Optional[ast.AstNode] = None,
    ) -> None:
        """Initialize."""
        self.name = name
        self.node = node

    def pretty_print(self) -> str:
        """Pretty print the symbol."""
        return pprint.pformat(vars(self), indent=2)

    def __repr__(self) -> str:
        """Representation for printing Symbols."""
        return f"{self.pretty_print()}"


T = TypeVar("T", bound=Symbol)


class TypedSymbol(Symbol):
    """Typed Symbol."""

    def __init__(
        self,
        name: str,
        typ: type,
        node: ast.AstNode,
        access: Optional[str] = None,
    ) -> None:
        """Initialize."""
        self.typ = typ
        self.access = access
        super().__init__(name, node)


class DefDeclSymbol(Symbol):
    """DefDecl Symbol."""

    def __init__(
        self,
        name: str,
        node: Optional[ast.AstNode] = None,
        other_node: Optional[ast.AstNode] = None,
        has_def: bool = False,
        has_decl: bool = False,
    ) -> None:
        """Initialize."""
        self.other_node = other_node
        self.has_def = has_def
        self.has_decl = has_decl
        super().__init__(name, node)


class SymbolTable:
    """Symbol Table.

    Functions as a stack by default, can create a separate table in pass
    if keeping track of multiple scopes is necessary.
    """

    def __init__(
        self,
        scope_name: str = "",
        parent: Optional["SymbolTable"] = None,
    ) -> None:
        """Initialize."""
        self.scope_name = scope_name
        self.parent = parent
        self.tab: dict = {}

    def lookup(self, name: str, deep: bool = True) -> Optional[T]:
        """Lookup a variable in the symbol table."""
        if name in self.tab:
            return self.tab[name]
        if deep and self.parent:
            return self.parent.lookup(name, deep)
        return None

    def set(self, sym: T, fresh_only: bool = False) -> bool:
        """Set a variable in the symbol table."""
        if fresh_only and sym.name in self.tab:
            return False
        self.tab[sym.name] = sym
        return True

    def update(self, sym: T, deep: bool = False) -> bool:
        """Set a variable in the symbol table."""
        if sym.name in self.tab:
            self.tab[sym.name] = T
            return True
        elif deep and self.parent:
            return self.parent.update(sym)
        return False

    def push(self, scope_name: str = "") -> "SymbolTable":
        """Push a new scope onto the symbol table."""
        return SymbolTable(scope_name, self)

    def pop(self) -> "SymbolTable":
        """Pop the current scope off the symbol table."""
        return self.parent if self.parent else self

    def pretty_print(self) -> str:
        """Pretty print the symbol table and return the result as a string."""
        output = f"Scope: {self.scope_name}\n"
        output += pprint.pformat(self.tab)
        if self.parent:
            output += f"\nParent Scope: {self.parent.scope_name}\n"
            output += self.parent.pretty_print()
        return output

    def __repr__(self) -> str:
        """Return the symbol table as a string."""
        return self.pretty_print()



"""Abstract class for IR Passes for Jac."""
from __future__ import annotations

import pprint
from typing import Optional, Union

from jaclang.jac.constant import Constants as Con, EdgeDir
from jaclang.jac.constant import Tokens as Tok


class AstNode:
    """Abstract syntax tree node for Jac."""

    def __init__(
        self,
        parent: Optional["AstNode"],
        mod_link: Optional[Module],
        kid: list,
        line: int,
    ) -> None:
        """Initialize ast."""
        self.parent = parent
        self.kid = kid if kid else []
        self.mod_link = mod_link
        self.line = line
        self._sub_node_tab: dict[type[AstNode], list[AstNode]] = {}
        self._typ: type = type(None)
        self.meta: dict = {}

    def to_dict(self) -> dict:
        """Return dict representation of node."""
        ret = {
            "node": str(type(self).__name__),
            "kid": [x.to_dict() for x in self.kid if x],
            "line": self.line,
        }
        if isinstance(self, Token):
            ret["name"] = self.name
            ret["value"] = self.value
        return ret

    def print(self, depth: Optional[int] = None) -> None:
        """Print ast."""
        pprint.PrettyPrinter(depth=depth).pprint(self.to_dict())


# AST Mid Level Node Types
# --------------------------
class Module(AstNode):
    """Whole Program node type for Jac Ast."""

    def __init__(
        self,
        name: str,
        doc: Token,
        body: Optional["Elements"],
        mod_path: str,
        rel_mod_path: str,
        is_imported: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize whole program node."""
        self.name = name
        self.doc = doc
        self.body = body
        self.mod_path = mod_path
        self.rel_mod_path = rel_mod_path
        self.is_imported = is_imported
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Elements(AstNode):
    """Elements node type for Jac Ast."""

    def __init__(
        self,
        elements: list[GlobalVars | Test | ModuleCode | Import | Architype | Ability],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize elements node."""
        self.elements = elements
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class OOPAccessNode(AstNode):
    """OOPAccessNode node type for Jac Ast."""

    def __init__(
        self,
        access: Optional[Token],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize OOPAccessible node."""
        self.access = access
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class GlobalVars(OOPAccessNode):
    """GlobalVars node type for Jac Ast."""

    def __init__(
        self,
        doc: Optional["Token"],
        access: Optional[Token],
        assignments: "AssignmentList",
        is_frozen: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize global var node."""
        self.doc = doc
        self.assignments = assignments
        self.is_frozen = is_frozen
        super().__init__(
            access=access, parent=parent, mod_link=mod_link, kid=kid, line=line
        )


class Test(AstNode):
    """Test node type for Jac Ast."""

    def __init__(
        self,
        name: Name,
        doc: Optional["Token"],
        description: Token,
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize test node."""
        self.doc = doc
        self.name = name
        self.description = description
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ModuleCode(AstNode):
    """Free mod code for Jac Ast."""

    def __init__(
        self,
        doc: Optional["Token"],
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize test node."""
        self.doc = doc
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Import(AstNode):
    """Import node type for Jac Ast."""

    def __init__(
        self,
        lang: Name,
        path: "ModulePath",
        alias: Optional[Name],
        items: Optional["ModuleItems"],
        is_absorb: bool,  # For includes
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
        sub_module: Optional["Module"] = None,
    ) -> None:
        """Initialize import node."""
        self.lang = lang
        self.path = path
        self.alias = alias
        self.items = items
        self.is_absorb = is_absorb
        self.sub_module = sub_module
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ModulePath(AstNode):
    """ModulePath node type for Jac Ast."""

    def __init__(
        self,
        path: list[Token],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize module path node."""
        self.path = path
        self.path_str = ".".join([p.value for p in path])
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ModuleItems(AstNode):
    """ModuleItems node type for Jac Ast."""

    def __init__(
        self,
        items: list["ModuleItem"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list["ModuleItem"],
        line: int,
    ) -> None:
        """Initialize module items node."""
        self.items = items
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ModuleItem(AstNode):
    """ModuleItem node type for Jac Ast."""

    def __init__(
        self,
        name: Name,
        alias: Optional[Token],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
        body: Optional[AstNode] = None,
    ) -> None:
        """Initialize module item node."""
        self.name = name
        self.alias = alias
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Architype(OOPAccessNode):
    """ObjectArch node type for Jac Ast."""

    def __init__(
        self,
        name: Name,
        arch_type: Token,
        doc: Optional[Token],
        decorators: Optional[Decorators],
        access: Optional[Token],
        base_classes: BaseClasses,
        body: Optional[ArchBlock],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize object arch node."""
        self.name = name
        self.arch_type = arch_type
        self.doc = doc
        self.decorators = decorators
        self.base_classes = base_classes
        self.body = body
        super().__init__(
            access=access, parent=parent, mod_link=mod_link, kid=kid, line=line
        )


class ArchDef(AstNode):
    """ArchDef node type for Jac Ast."""

    def __init__(
        self,
        doc: Optional[Token],
        mod: Optional[DottedNameList],
        arch: ArchRef,
        body: ArchBlock,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize arch def node."""
        self.doc = doc
        self.mod = mod
        self.arch = arch
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Decorators(AstNode):
    """Decorators node type for Jac Ast."""

    def __init__(
        self,
        calls: list[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize decorators node."""
        self.calls = calls
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class BaseClasses(AstNode):
    """BaseArch node type for Jac Ast."""

    def __init__(
        self,
        base_classes: list[DottedNameList],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize base classes node."""
        self.base_classes = base_classes
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Ability(OOPAccessNode):
    """Ability node type for Jac Ast."""

    def __init__(
        self,
        name_ref: Name | SpecialVarRef | ArchRef,
        is_func: bool,
        is_async: bool,
        is_static: bool,
        doc: Optional[Token],
        decorators: Optional[Decorators],
        access: Optional[Token],
        signature: Optional[FuncSignature | TypeSpec | EventSignature],
        body: Optional[CodeBlock],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
        arch_attached: Optional[ArchBlock] = None,
    ) -> None:
        """Initialize func arch node."""
        self.name_ref = name_ref
        self.is_func = is_func
        self.is_async = is_async
        self.is_static = is_static
        self.doc = doc
        self.decorators = decorators
        self.signature = signature
        self.body = body
        self.arch_attached = arch_attached
        super().__init__(
            access=access, parent=parent, mod_link=mod_link, kid=kid, line=line
        )

    def py_resolve_name(self) -> str:
        """Resolve name."""
        if isinstance(self.name_ref, Name):
            return self.name_ref.value
        elif isinstance(self.name_ref, (SpecialVarRef, ArchRef)):
            return self.name_ref.py_resolve_name()
        else:
            raise NotImplementedError


class AbilityDef(AstNode):
    """AbilityDef node type for Jac Ast."""

    def __init__(
        self,
        doc: Optional[Token],
        target: Optional[DottedNameList],
        ability: ArchRef,
        signature: FuncSignature | EventSignature,
        body: CodeBlock,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize ability def node."""
        self.doc = doc
        self.target = target
        self.ability = ability
        self.signature = signature
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class EventSignature(AstNode):
    """EventSignature node type for Jac Ast."""

    def __init__(
        self,
        event: Token,
        arch_tag_info: Optional[TypeSpecList],
        return_type: Optional["TypeSpec"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize event signature node."""
        self.event = event
        self.arch_tag_info = arch_tag_info
        self.return_type = return_type
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class DottedNameList(AstNode):
    """DottedNameList node type for Jac Ast."""

    def __init__(
        self,
        names: list[Token | SpecialVarRef | ArchRef | Name],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize name list node."""
        self.names = names
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class FuncSignature(AstNode):
    """FuncSignature node type for Jac Ast."""

    def __init__(
        self,
        params: Optional["FuncParams"],
        return_type: Optional["TypeSpec"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize method signature node."""
        self.params = params
        self.return_type = return_type
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class FuncParams(AstNode):
    """ArchBlock node type for Jac Ast."""

    def __init__(
        self,
        params: list["ParamVar"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize method params node."""
        self.params = params
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ParamVar(AstNode):
    """ParamVar node type for Jac Ast."""

    def __init__(
        self,
        name: Name,
        unpack: Optional[Token],
        type_tag: "TypeSpec",
        value: Optional[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize param var node."""
        self.name = name
        self.unpack = unpack
        self.type_tag = type_tag
        self.value = value
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Enum(OOPAccessNode):
    """Enum node type for Jac Ast."""

    def __init__(
        self,
        name: Name,
        doc: Optional[Token],
        decorators: Optional["Decorators"],
        access: Optional[Token],
        base_classes: "BaseClasses",
        body: Optional["EnumBlock"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize object arch node."""
        self.name = name
        self.doc = doc
        self.decorators = decorators
        self.base_classes = base_classes
        self.body = body
        super().__init__(
            access=access, parent=parent, mod_link=mod_link, kid=kid, line=line
        )


class EnumDef(AstNode):
    """EnumDef node type for Jac Ast."""

    def __init__(
        self,
        doc: Optional[Token],
        enum: ArchRef,
        mod: Optional[DottedNameList],
        body: EnumBlock,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize arch def node."""
        self.enum = enum
        self.doc = doc
        self.mod = mod
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class EnumBlock(AstNode):
    """EnumBlock node type for Jac Ast."""

    def __init__(
        self,
        stmts: list["Name|Assignment"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize enum block node."""
        self.stmts = stmts
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ArchBlock(AstNode):
    """ArchBlock node type for Jac Ast."""

    def __init__(
        self,
        members: list["ArchHas | Ability"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize arch block node."""
        self.members = members
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ArchHas(OOPAccessNode):
    """HasStmt node type for Jac Ast."""

    counter: int = 1

    def __init__(
        self,
        doc: Optional[Token],
        is_static: bool,
        access: Optional[Token],
        vars: "HasVarList",
        is_frozen: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize has statement node."""
        self.doc = doc
        self.is_static = is_static
        self.vars = vars
        self.is_frozen = is_frozen
        self.h_id = ArchHas.counter
        ArchHas.counter += 1
        super().__init__(
            access=access, parent=parent, mod_link=mod_link, kid=kid, line=line
        )


class HasVarList(AstNode):
    """HasVarList node type for Jac Ast."""

    def __init__(
        self,
        vars: list["HasVar"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize has var list node."""
        self.vars = vars
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class HasVar(AstNode):
    """HasVar node type for Jac Ast."""

    def __init__(
        self,
        name: Name,
        type_tag: "TypeSpec",
        value: Optional[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize has var node."""
        self.name = name
        self.type_tag = type_tag
        self.value = value
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class TypeSpecList(AstNode):
    """TypeSpecList node type for Jac Ast."""

    def __init__(
        self,
        types: list[TypeSpec],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize type list node."""
        self.types = types
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class TypeSpec(AstNode):
    """TypeSpec node type for Jac Ast."""

    def __init__(
        self,
        spec_type: Token | DottedNameList,
        list_nest: TypeSpec,  # needed for lists
        dict_nest: TypeSpec,  # needed for dicts, uses list_nest as key
        null_ok: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize type spec node."""
        self.spec_type = spec_type
        self.list_nest = list_nest
        self.dict_nest = dict_nest
        self.null_ok = null_ok
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class CodeBlock(AstNode):
    """CodeBlock node type for Jac Ast."""

    def __init__(
        self,
        stmts: list[StmtType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize code block node."""
        self.stmts = stmts
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class TypedCtxBlock(AstNode):
    """TypedCtxBlock node type for Jac Ast."""

    def __init__(
        self,
        type_ctx: TypeSpecList,
        body: CodeBlock,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize typed context block node."""
        self.type_ctx = type_ctx
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class IfStmt(AstNode):
    """IfStmt node type for Jac Ast."""

    def __init__(
        self,
        condition: ExprType,
        body: "CodeBlock",
        elseifs: Optional["ElseIfs"],
        else_body: Optional["ElseStmt"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize if statement node."""
        self.condition = condition
        self.body = body
        self.elseifs = elseifs
        self.else_body = else_body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ElseIfs(AstNode):
    """ElseIfs node type for Jac Ast."""

    def __init__(
        self,
        elseifs: list["IfStmt"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list["IfStmt"],
        line: int,
    ) -> None:
        """Initialize elseifs node."""
        self.elseifs = elseifs
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ElseStmt(AstNode):
    """Else node type for Jac Ast."""

    def __init__(
        self,
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize else node."""
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class TryStmt(AstNode):
    """TryStmt node type for Jac Ast."""

    def __init__(
        self,
        body: "CodeBlock",
        excepts: Optional["ExceptList"],
        finally_body: Optional["FinallyStmt"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize try statement node."""
        self.body = body
        self.excepts = excepts
        self.finally_body = finally_body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ExceptList(AstNode):
    """ExceptList node type for Jac Ast."""

    def __init__(
        self,
        excepts: list["Except"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize excepts node."""
        self.excepts = excepts
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Except(AstNode):
    """Except node type for Jac Ast."""

    def __init__(
        self,
        ex_type: ExprType,
        name: Optional[Token],
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize except node."""
        self.ex_type = ex_type
        self.name = name
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class FinallyStmt(AstNode):
    """FinallyStmt node type for Jac Ast."""

    def __init__(
        self,
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize finally statement node."""
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class IterForStmt(AstNode):
    """IterFor node type for Jac Ast."""

    def __init__(
        self,
        iter: "Assignment",
        condition: ExprType,
        count_by: ExprType,
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize iter for node."""
        self.iter = iter
        self.condition = condition
        self.count_by = count_by
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class InForStmt(AstNode):
    """InFor node type for Jac Ast."""

    def __init__(
        self,
        name_list: NameList,
        collection: ExprType,
        body: CodeBlock,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize in for node."""
        self.name_list = name_list
        self.collection = collection
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class NameList(AstNode):
    """NameList node type for Jac Ast."""

    def __init__(
        self,
        names: list[Name],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize name list node."""
        self.names = names
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class WhileStmt(AstNode):
    """WhileStmt node type for Jac Ast."""

    def __init__(
        self,
        condition: ExprType,
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize while statement node."""
        self.condition = condition
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class WithStmt(AstNode):
    """WithStmt node type for Jac Ast."""

    def __init__(
        self,
        exprs: "ExprAsItemList",
        body: "CodeBlock",
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize with statement node."""
        self.exprs = exprs
        self.body = body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ExprAsItemList(AstNode):
    """ExprAsItemList node type for Jac Ast."""

    def __init__(
        self,
        items: list["ExprAsItem"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list["ExprAsItem"],
        line: int,
    ) -> None:
        """Initialize module items node."""
        self.items = items
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ExprAsItem(AstNode):
    """ExprAsItem node type for Jac Ast."""

    def __init__(
        self,
        expr: ExprType,
        alias: Optional[Name],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize module item node."""
        self.expr = expr
        self.alias = alias
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class RaiseStmt(AstNode):
    """RaiseStmt node type for Jac Ast."""

    def __init__(
        self,
        cause: Optional[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize raise statement node."""
        self.cause = cause
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class AssertStmt(AstNode):
    """AssertStmt node type for Jac Ast."""

    def __init__(
        self,
        condition: ExprType,
        error_msg: Optional[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize assert statement node."""
        self.condition = condition
        self.error_msg = error_msg
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class CtrlStmt(AstNode):
    """CtrlStmt node type for Jac Ast."""

    def __init__(
        self,
        ctrl: Token,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize control statement node."""
        self.ctrl = ctrl
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class DeleteStmt(AstNode):
    """DeleteStmt node type for Jac Ast."""

    def __init__(
        self,
        target: ExprType,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize delete statement node."""
        self.target = target
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ReportStmt(AstNode):
    """ReportStmt node type for Jac Ast."""

    def __init__(
        self,
        expr: ExprType,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize report statement node."""
        self.expr = expr
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ReturnStmt(AstNode):
    """ReturnStmt node type for Jac Ast."""

    def __init__(
        self,
        expr: Optional[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize return statement node."""
        self.expr = expr
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class YieldStmt(AstNode):
    """YieldStmt node type for Jac Ast."""

    def __init__(
        self,
        expr: Optional[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize yeild statement node."""
        self.expr = expr
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class IgnoreStmt(AstNode):
    """IgnoreStmt node type for Jac Ast."""

    def __init__(
        self,
        target: ExprType,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
        from_walker: bool = False,
    ) -> None:
        """Initialize ignore statement node."""
        self.target = target
        self.from_walker = from_walker
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class VisitStmt(AstNode):
    """VisitStmt node type for Jac Ast."""

    def __init__(
        self,
        vis_type: Optional[Token],
        target: ExprType,
        else_body: Optional["ElseStmt"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
        from_walker: bool = False,
    ) -> None:
        """Initialize visit statement node."""
        self.vis_type = vis_type
        self.target = target
        self.else_body = else_body
        self.from_walker = from_walker
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class RevisitStmt(AstNode):
    """ReVisitStmt node type for Jac Ast."""

    def __init__(
        self,
        hops: Optional[ExprType],
        else_body: Optional["ElseStmt"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize revisit statement node."""
        self.hops = hops
        self.else_body = else_body
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class DisengageStmt(AstNode):
    """DisengageStmt node type for Jac Ast."""

    def __init__(
        self,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
        from_walker: bool = False,
    ) -> None:
        """Initialize disengage statement node."""
        self.from_walker = from_walker
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class AwaitStmt(AstNode):
    """AwaitStmt node type for Jac Ast."""

    def __init__(
        self,
        target: ExprType,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize sync statement node."""
        self.target = target
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Assignment(AstNode):
    """Assignment node type for Jac Ast."""

    def __init__(
        self,
        is_static: bool,
        target: "AtomType",
        value: ExprType,
        mutable: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize assignment node."""
        self.is_static = is_static
        self.target = target
        self.value = value
        self.mutable = mutable
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class BinaryExpr(AstNode):
    """ExprBinary node type for Jac Ast."""

    def __init__(
        self,
        left: ExprType,
        right: ExprType,
        op: Token | DisconnectOp | ConnectOp,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize binary expression node."""
        self.left = left
        self.right = right
        self.op = op
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class IfElseExpr(AstNode):
    """ExprIfElse node type for Jac Ast."""

    def __init__(
        self,
        condition: "BinaryExpr | IfElseExpr",
        value: ExprType,
        else_value: ExprType,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize if else expression node."""
        self.condition = condition
        self.value = value
        self.else_value = else_value
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class UnaryExpr(AstNode):
    """ExprUnary node type for Jac Ast."""

    def __init__(
        self,
        operand: ExprType,
        op: Token,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize unary expression node."""
        self.operand = operand
        self.op = op
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class UnpackExpr(AstNode):
    """ExprUnpack node type for Jac Ast."""

    def __init__(
        self,
        target: ExprType,
        is_dict: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize unpack expression node."""
        self.target = target
        self.is_dict = is_dict
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class MultiString(AstNode):
    """ExprMultiString node type for Jac Ast."""

    def __init__(
        self,
        strings: list["Token | FString"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize multi string expression node."""
        self.strings = strings
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ExprList(AstNode):
    """ExprList node type for Jac Ast."""

    def __init__(
        self,
        values: list[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize expr value node."""
        self.values = values
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ListVal(ExprList):
    """ListVal node type for Jac Ast."""


class SetVal(ExprList):
    """SetVal node type for Jac Ast."""


class TupleVal(AstNode):
    """TupleVal node type for Jac Ast."""

    def __init__(
        self,
        first_expr: Optional[ExprType],
        exprs: Optional[ExprList],
        assigns: Optional[AssignmentList],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize tuple value node."""
        self.first_expr = first_expr
        self.exprs = exprs
        self.assigns = assigns
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class DictVal(AstNode):
    """ExprDict node type for Jac Ast."""

    def __init__(
        self,
        kv_pairs: list["KVPair"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize dict expression node."""
        self.kv_pairs = kv_pairs
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class InnerCompr(AstNode):
    """ListCompr node type for Jac Ast."""

    def __init__(
        self,
        out_expr: ExprType,
        name_list: NameList,
        collection: ExprType,
        conditional: Optional[ExprType],
        is_list: bool,
        is_gen: bool,
        is_set: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize comprehension expression node."""
        self.out_expr = out_expr
        self.name_list = name_list
        self.collection = collection
        self.conditional = conditional
        self.is_list = is_list
        self.is_gen = is_gen
        self.is_set = is_set
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class DictCompr(AstNode):
    """DictCompr node type for Jac Ast."""

    def __init__(
        self,
        outk_expr: ExprType,
        outv_expr: ExprType,
        name_list: NameList,
        collection: ExprType,
        conditional: Optional[ExprType],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize comprehension expression node."""
        self.outk_expr = outk_expr
        self.outv_expr = outv_expr
        self.name_list = name_list
        self.collection = collection
        self.conditional = conditional
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class KVPair(AstNode):
    """ExprKVPair node type for Jac Ast."""

    def __init__(
        self,
        key: ExprType,
        value: ExprType,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize key value pair expression node."""
        self.key = key
        self.value = value
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class AtomTrailer(AstNode):
    """AtomTrailer node type for Jac Ast."""

    def __init__(
        self,
        target: "AtomType",
        right: "IndexSlice | ArchRef | Token",
        null_ok: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize atom trailer expression node."""
        self.target = target
        self.right = right
        self.null_ok = null_ok
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class FuncCall(AstNode):
    """FuncCall node type for Jac Ast."""

    def __init__(
        self,
        target: "AtomType",
        params: Optional["ParamList"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize function call expression node."""
        self.target = target
        self.params = params
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ParamList(AstNode):
    """ParamList node type for Jac Ast."""

    def __init__(
        self,
        p_args: Optional[ExprList],
        p_kwargs: Optional["AssignmentList"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize parameter list expression node."""
        self.p_args = p_args
        self.p_kwargs = p_kwargs
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class AssignmentList(AstNode):
    """AssignmentList node type for Jac Ast."""

    def __init__(
        self,
        values: list["Assignment"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize expr value node."""
        self.values = values
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class IndexSlice(AstNode):
    """IndexSlice node type for Jac Ast."""

    def __init__(
        self,
        start: Optional[ExprType],
        stop: Optional[ExprType],
        is_range: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize index slice expression node."""
        self.start = start
        self.stop = stop
        self.is_range = is_range
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class ArchRef(AstNode):
    """GlobalRef node type for Jac Ast."""

    def __init__(
        self,
        name_ref: Name | SpecialVarRef,
        arch: Token,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize global reference expression node."""
        self.name_ref = name_ref
        self.arch = arch
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)

    def py_resolve_name(self) -> str:
        """Resolve name."""
        if isinstance(self.name_ref, Name):
            return self.name_ref.value
        elif isinstance(self.name_ref, SpecialVarRef):
            return self.name_ref.py_resolve_name()
        else:
            raise NotImplementedError


class SpecialVarRef(AstNode):
    """HereRef node type for Jac Ast."""

    def __init__(
        self,
        var: Token,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize special var reference expression node."""
        self.var = var
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)

    def py_resolve_name(self) -> str:
        """Resolve name."""
        if self.var.name == Tok.SELF_OP:
            return "self"
        elif self.var.name == Tok.SUPER_OP:
            return "super()"
        elif self.var.name == Tok.ROOT_OP:
            return Con.ROOT
        elif self.var.name == Tok.HERE_OP:
            return Con.HERE
        elif self.var.name == Tok.INIT_OP:
            return "__init__"
        else:
            raise NotImplementedError("ICE: Special var reference not implemented")


class EdgeOpRef(AstNode):
    """EdgeOpRef node type for Jac Ast."""

    def __init__(
        self,
        filter_type: Optional[ExprType],
        filter_cond: Optional[FilterCompr],
        edge_dir: EdgeDir,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
        from_walker: bool = False,
    ) -> None:
        """Initialize edge op reference expression node."""
        self.filter_type = filter_type
        self.filter_cond = filter_cond
        self.edge_dir = edge_dir
        self.from_walker = from_walker
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class DisconnectOp(EdgeOpRef):
    """DisconnectOpRef node type for Jac Ast."""


class ConnectOp(AstNode):
    """ConnectOpRef node type for Jac Ast."""

    def __init__(
        self,
        conn_type: Optional[ExprType],
        conn_assign: Optional[AssignmentList],
        edge_dir: EdgeDir,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize connect op reference expression node."""
        self.conn_type = conn_type
        self.conn_assign = conn_assign
        self.edge_dir = edge_dir
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class FilterCompr(AstNode):
    """FilterCtx node type for Jac Ast."""

    def __init__(
        self,
        compares: list[BinaryExpr],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize filter_cond context expression node."""
        self.compares = compares
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class FString(AstNode):
    """FString node type for Jac Ast."""

    def __init__(
        self,
        parts: list["Token | ExprType"],
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize fstring expression node."""
        self.parts = parts
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


AtomType = Union[
    MultiString,
    ListVal,
    TupleVal,
    SetVal,
    DictVal,
    InnerCompr,
    DictCompr,
    AtomTrailer,
    EdgeOpRef,
    FilterCompr,
]

ExprType = Union[
    UnaryExpr,
    BinaryExpr,
    IfElseExpr,
    UnpackExpr,
    AtomType,
]


StmtType = Union[
    Import,
    Architype,
    Ability,
    Assignment,
    ExprType,
    IfStmt,
    TryStmt,
    IterForStmt,
    InForStmt,
    WhileStmt,
    WithStmt,
    RaiseStmt,
    AssertStmt,
    CtrlStmt,
    DeleteStmt,
    ReportStmt,
    ReturnStmt,
    YieldStmt,
    AwaitStmt,
    DisengageStmt,
    RevisitStmt,
    VisitStmt,
    IgnoreStmt,
]

# AST Parse-Tree Node Types
# --------------------------


class Parse(AstNode):
    """Parse node type for Jac Ast."""

    def __init__(
        self,
        name: str,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize parse."""
        self.name = name
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Token(AstNode):
    """Token node type for Jac Ast."""

    def __init__(
        self,
        name: str,
        value: str,
        col_start: int,
        col_end: int,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize token."""
        self.name = name
        self.value = value
        self.col_start = col_start
        self.col_end = col_end
        super().__init__(parent=parent, mod_link=mod_link, kid=kid, line=line)


class Name(Token):
    """Name node type for Jac Ast."""

    def __init__(
        self,
        name: str,
        value: str,
        col_start: int,
        col_end: int,
        already_declared: bool,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize name."""
        self.already_declared = already_declared
        super().__init__(
            name=name,
            value=value,
            col_start=col_start,
            col_end=col_end,
            parent=parent,
            mod_link=mod_link,
            kid=kid,
            line=line,
        )


class Constant(Token):
    """Constant node type for Jac Ast."""

    def __init__(
        self,
        name: str,
        value: str,
        col_start: int,
        col_end: int,
        typ: type,
        parent: Optional[AstNode],
        mod_link: Optional[Module],
        kid: list[AstNode],
        line: int,
    ) -> None:
        """Initialize constant."""
        super().__init__(
            name=name,
            value=value,
            col_start=col_start,
            col_end=col_end,
            parent=parent,
            mod_link=mod_link,
            kid=kid,
            line=line,
        )
        self._typ = typ


# Utiliiy functions
# -----------------


def replace_node(node: AstNode, new_node: Optional[AstNode]) -> AstNode | None:
    """Replace node with new_node."""
    if node.parent:
        node.parent.kid[node.parent.kid.index(node)] = new_node
    if new_node:
        new_node.parent = node.parent
        for i in new_node.kid:
            if i:
                i.parent = new_node
    return new_node


def append_node(node: AstNode, new_node: Optional[AstNode]) -> AstNode | None:
    """Replace node with new_node."""
    node.kid.append(new_node)
    if new_node:
        new_node.parent = node
    return new_node



"""Test case utils for Jaseci."""

import inspect
import os
from abc import ABC, abstractmethod
from unittest import TestCase as _TestCase


class TestCase(_TestCase):
    """Base test case for Jaseci."""

    def setUp(self) -> None:
        """Set up test case."""
        return super().setUp()

    def tearDown(self) -> None:
        """Tear down test case."""
        return super().tearDown()

    def load_fixture(self, fixture: str) -> str:
        """Load fixture from fixtures directory."""
        frame = inspect.currentframe()
        if frame is None or frame.f_back is None:
            raise ValueError("Unable to get the previous stack frame.")
        module = inspect.getmodule(frame.f_back)
        if module is None or module.__file__ is None:
            raise ValueError("Unable to determine the file of the module.")
        fixture_src = module.__file__
        fixture_path = os.path.join(os.path.dirname(fixture_src), "fixtures", fixture)
        with open(fixture_path, "r") as f:
            return f.read()

    def file_to_str(self, file_path: str) -> str:
        """Load fixture from fixtures directory."""
        with open(file_path, "r") as f:
            return f.read()

    def fixture_abs_path(self, fixture: str) -> str:
        """Get absolute path of a fixture from fixtures directory."""
        frame = inspect.currentframe()
        if frame is None or frame.f_back is None:
            raise ValueError("Unable to get the previous stack frame.")
        module = inspect.getmodule(frame.f_back)
        if module is None or module.__file__ is None:
            raise ValueError("Unable to determine the file of the module.")
        fixture_src = module.__file__
        file_path = os.path.join(os.path.dirname(fixture_src), "fixtures", fixture)
        return os.path.abspath(file_path)


class TestCaseMicroSuite(ABC, TestCase):
    """Base test case for Jaseci."""

    @classmethod
    def self_attach_micro_tests(cls) -> None:
        """Attach micro tests."""
        directory = os.path.dirname(__file__) + "/../../examples/micro"
        for filename in os.listdir(directory):
            if os.path.isfile(os.path.join(directory, filename)) and filename.endswith(
                ".jac"
            ):
                method_name = f"test_micro_{filename.replace('.jac', '')}"
                file_path = os.path.join(directory, filename)
                setattr(
                    cls, method_name, lambda self, f=file_path: self.micro_suite_test(f)
                )

        directory = os.path.dirname(__file__) + "/../../examples/guess_game"
        for filename in os.listdir(directory):
            if os.path.isfile(os.path.join(directory, filename)) and filename.endswith(
                ".jac"
            ):
                method_name = f"test_gg_{filename.replace('.jac', '')}"
                file_path = os.path.join(directory, filename)
                setattr(
                    cls, method_name, lambda self, f=file_path: self.micro_suite_test(f)
                )

        def test_micro_jac_files_fully_tested(self) -> None:  # noqa: ANN001
            """Test that all micro jac files are fully tested."""
            self.directory = os.path.dirname(__file__) + "/../../examples/micro"
            for filename in os.listdir(self.directory):
                if os.path.isfile(os.path.join(self.directory, filename)):
                    method_name = f"test_micro_{filename.replace('.jac', '')}"
                    self.assertIn(method_name, dir(self))

        cls.test_micro_jac_files_fully_tested = test_micro_jac_files_fully_tested

    @abstractmethod
    def micro_suite_test(self, filename: str) -> None:
        """Test micro jac file."""
        pass



"""Connect Decls and Defs in AST."""
import jaclang.jac.absyntree as ast
from jaclang.jac.passes import Pass
from jaclang.jac.passes.blue import SubNodeTabPass
from jaclang.jac.sym_table import DefDeclSymbol, SymbolTable


class DeclDefMatchPass(Pass, SymbolTable):
    """Decls and Def matching pass."""

    def before_pass(self) -> None:
        """Initialize pass."""
        self.sym_tab = SymbolTable(scope_name="global")

    def after_pass(self) -> None:
        """Rebuild sub node table."""
        self.ir = SubNodeTabPass(mod_path=self.mod_path, input_ir=self.ir).ir

    def exit_global_vars(self, node: ast.GlobalVars) -> None:
        """Sub objects.

        doc: Optional[DocString],
        access: Optional[Token],
        assignments: AssignmentList,
        """
        for i in self.get_all_sub_nodes(node, ast.Assignment):
            if not isinstance(i.target, ast.Name):
                self.ice("Only name targets should be possible to in global vars.")
            else:
                decl = self.sym_tab.lookup(i.target.value)
                if decl:
                    if decl.has_def:
                        self.error(f"Name {i.target.value} already bound.")
                    else:
                        decl.has_def = True
                        decl.other_node = i
                        decl.node.body = i  # TODO: I dont think this line makes sense
                        self.sym_tab.set(decl)

    def exit_test(self, node: ast.Test) -> None:
        """Sub objects.

        name: Name,
        doc: Optional[DocString],
        description: Token,
        body: CodeBlock,
        """

    def enter_import(self, node: ast.Import) -> None:
        """Sub objects.

        lang: Name,
        path: ModulePath,
        alias: Optional[Name],
        items: Optional[ModuleItems],
        is_absorb: bool,
        sub_module: Optional[Module],
        """
        if not node.is_absorb:
            self.sym_tab = self.sym_tab.push(node.path.path_str)

    def exit_import(self, node: ast.Import) -> None:
        """Sub objects.

        lang: Name,
        path: ModulePath,
        alias: Optional[Name],
        items: Optional[ModuleItems],
        is_absorb: bool,
        sub_module: Optional[Module],
        """
        if not node.is_absorb and not self.sym_tab.parent:
            self.ice("Import should have a parent sym_table scope.")
        elif not node.is_absorb:
            self.sym_tab = self.sym_tab.pop()
        if node.items:  # now treat imported items as global
            for i in node.items.items:
                name = i.alias if i.alias else i.name
                decl = self.sym_tab.lookup(name.value)
                if not decl:
                    self.sym_tab.set(
                        DefDeclSymbol(name=name.value, node=i, has_def=True)
                    )

    def exit_module_item(self, node: ast.ModuleItem) -> None:
        """Sub objects.

        name: Name,
        alias: Optional[Token],
        body: Optional[AstNode],
        """
        if not self.sym_tab.lookup(node.name.value):
            self.sym_tab.set(
                DefDeclSymbol(name=node.name.value, node=node, has_decl=True)
            )

    def exit_architype(self, node: ast.Architype) -> None:
        """Sub objects.

        name: Name,
        arch_type: Token,
        doc: Optional[DocString],
        decorators: Optional[Decorators],
        access: Optional[Token],
        base_classes: BaseClasses,
        body: Optional[ArchBlock],
        """
        # if no body, check for def
        #    if no def, register as decl
        # if complete register as def
        # nota: can allow static overriding perhaps?
        # note: if arch has not body ok, imports body is the arch itself

    def exit_arch_def(self, node: ast.ArchDef) -> None:
        """Sub objects.

        doc: Optional[DocString],
        mod: Optional[DottedNameList],
        arch: ObjectRef | NodeRef | EdgeRef | WalkerRef,
        body: ArchBlock,
        """

    def enter_ability(self, node: ast.Ability) -> None:
        """Sub objects.

        name_ref: Name | SpecialVarRef | ArchRef,
        is_func: bool,
        is_async: bool,
        is_static: bool,
        doc: Optional[Token],
        decorators: Optional["Decorators"],
        access: Optional[Token],
        signature: Optional["FuncSignature | TypeSpec | EventSignature"],
        body: Optional["CodeBlock"],

        arch_attached: Optional["ArchBlock"] = None,
        """
        self.sym_tab = self.sym_tab.push(node.py_resolve_name())

    def exit_ability(self, node: ast.Ability) -> None:
        """Sub objects.

        name: Name,
        is_func: bool,
        doc: Optional[DocString],
        decorators: Optional["Decorators"],
        access: Optional[Token],
        signature: "FuncSignature | TypeSpec | EventSignature",
        body: Optional["CodeBlock"],
        arch_attached: Optional["ArchBlock"] = None,
        """
        self.sym_tab = self.sym_tab.pop()
        ability_name = node.py_resolve_name()
        name = (
            f"{node.arch_attached.parent.name.value}.{ability_name}"
            if node.arch_attached
            and isinstance(node.arch_attached.parent, ast.Architype)
            else ability_name
        )
        decl = self.sym_tab.lookup(name)
        if decl and decl.has_decl:
            self.error(
                f"Ability bound with name {name} already defined on "
                f"Line {decl.node.line} in {decl.node.mod_link.rel_mod_path}."
            )
        elif decl and decl.has_def:
            decl.has_decl = True
            decl.node = node
            decl.node.body = (
                decl.other_node.body
                if isinstance(decl.other_node, ast.AbilityDef)
                else self.ice("Expected node of type AbilityDef in symbol table.")
            )
            ast.append_node(decl.node, decl.other_node)
            self.sym_tab.set(decl)
        else:
            decl = DefDeclSymbol(name=name, node=node, has_decl=True)
            if node.body:
                decl.has_def = True
                decl.other_node = node
            self.sym_tab.set(decl)

    def enter_ability_def(self, node: ast.AbilityDef) -> None:
        """Sub objects.

        doc: Optional[DocString],
        target: Optional["DottedNameList"],
        ability: "ArchRef",
        signature: "FuncSignature | EventSignature",
        body: "CodeBlock",
        """
        self.sym_tab = self.sym_tab.push(node.ability.py_resolve_name())

    def exit_ability_def(self, node: ast.AbilityDef) -> None:
        """Sub objects.

        doc: Optional[DocString],
        target: Optional["DottedNameList"],
        ability: "ArchRef",
        signature: "FuncSignature | EventSignature",
        body: "CodeBlock",
        """
        self.sym_tab = self.sym_tab.pop()
        name = node.ability.py_resolve_name()
        if node.target:
            owner = node.target.names[-1]
            if not isinstance(owner, ast.ArchRef):
                self.error("Expected reference to Architype!")
                owner = ""
            else:
                owner = owner.py_resolve_name()
            name = f"{owner}.{name}"
        decl = self.sym_tab.lookup(name)
        if decl and decl.has_def:
            self.error(
                f"Ability bound with name {name} already defined on "
                f"Line {decl.other_node.line} in {decl.other_node.mod_link.rel_mod_path}."
            )
        elif decl and decl.has_decl:
            decl.has_def = True
            decl.other_node = node
            decl.node.body = decl.other_node.body
            ast.append_node(decl.node, decl.other_node)
            self.sym_tab.set(decl)
        else:
            self.sym_tab.set(DefDeclSymbol(name=name, other_node=node, has_def=True))

    def enter_arch_block(self, node: ast.ArchBlock) -> None:
        """Sub objects.

        members: list['ArchHas | Ability'],
        """
        # Tags all function signatures whether method style or not
        for i in self.get_all_sub_nodes(node, ast.Ability):
            i.arch_attached = node

    def exit_enum(self, node: ast.Enum) -> None:
        """Sub objects.

        name: Name,
        doc: Optional[DocString],
        decorators: Optional[Decorators],
        access: Optional[Token],
        base_classes: BaseClasses,
        body: Optional[EnumBlock],
        """
        name = node.name.value
        decl = self.sym_tab.lookup(name)
        if decl and decl.has_decl:
            self.error(
                f"Enum bound with name {name} already defined on Line {decl.node.line}."
            )
        elif decl and decl.has_def:
            decl.has_decl = True
            decl.node = node
            decl.node.body = (
                decl.other_node.body
                if isinstance(decl.other_node, ast.EnumDef)
                else self.ice("Expected node of type EnumDef in symbol table.")
            )
            ast.append_node(decl.node, decl.other_node)
            self.sym_tab.set(decl)
        else:
            decl = DefDeclSymbol(name=name, node=node, has_decl=True)
            if node.body:
                decl.has_def = True
                decl.other_node = node
            self.sym_tab.set(decl)

    def exit_enum_def(self, node: ast.EnumDef) -> None:
        """Sub objects.

        doc: Optional[Token],
        enum: ArchRef,
        mod: Optional[DottedNameList],
        body: EnumBlock,
        """
        name = node.enum.py_resolve_name()
        decl = self.sym_tab.lookup(name)
        if decl and decl.has_def:
            self.error(
                f"Enum bound with name {name} already defined on Line {decl.other_node.line}."
            )
        elif decl and decl.has_decl:
            decl.has_def = True
            decl.other_node = node
            decl.node.body = decl.other_node.body
            ast.append_node(decl.node, decl.other_node)
            self.sym_tab.set(decl)
        else:
            self.sym_tab.set(DefDeclSymbol(name=name, other_node=node, has_def=True))



"""Collection of passes for Jac IR."""
from .ast_build_pass import AstBuildPass  # noqa: I100
from .sub_node_tab_pass import SubNodeTabPass
from .import_pass import ImportPass  # noqa: I100
from .decl_def_match_pass import DeclDefMatchPass  # noqa: I100
from .type_analyze_pass import TypeAnalyzePass
from .blue_pygen_pass import BluePygenPass  # noqa: I100

pass_schedule = [
    AstBuildPass,
    SubNodeTabPass,
    ImportPass,
    DeclDefMatchPass,
    TypeAnalyzePass,
    BluePygenPass,
]

__all__ = [
    "AstBuildPass",
    "SubNodeTabPass",
    "ImportPass",
    "DeclDefMatchPass",
    "TypeAnalyzePass",
    "BluePygenPass",
]



"""Test pass module."""
from jaclang.jac.passes.blue import ImportPass
from jaclang.jac.transpiler import jac_file_to_pass
from jaclang.utils.test import TestCase


class ImportPassPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_pygen_jac_cli(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(self.fixture_abs_path("base.jac"), "", ImportPass)
        self.assertFalse(state.errors_had)
        self.assertIn("56", str(state.ir.to_dict()))



# type: ignore
"""Python Like F-String Parser."""
from jaclang.utils.sly.lex import Lexer
from jaclang.utils.sly.yacc import Parser, YaccProduction

_ = None  # For flake8 linting and sly compatibility


class FStringLexer(Lexer):
    """Python Like F-String Lexer."""

    tokens = {
        "STRING_START",
        "STRING_END",
        "EXPR_START",
        "EXPR_END",
        "PIECE",
    }
    # ignore = " \t"

    # Tokens

    STRING_START = r"f\""
    STRING_END = r"\""
    PIECE = r"[^\{\}\"]+|\{\{|\}\}"
    EXPR_START = r"(?<!\{)\{(?!\{)"
    EXPR_END = r"(?<!\})\}(?!\})"


class FStringParser(Parser):
    """Python Like F-String Parser."""

    tokens = FStringLexer.tokens
    start = "fstring"

    @_("STRING_START fstr_parts STRING_END")
    def fstring(self, p: YaccProduction) -> YaccProduction:
        """Start rule for fstring."""
        return p

    @_(
        "fstr_parts fstr_expr",
        "fstr_parts PIECE",
        "PIECE",
        "fstr_expr",
        "fstring",
    )
    def fstr_parts(self, p: YaccProduction) -> YaccProduction:
        """Parts of the string both in string and in expression."""
        return p

    @_("EXPR_START fstr_parts EXPR_END")
    def fstr_expr(self, p: YaccProduction) -> YaccProduction:
        """Expressions rule."""
        return p


if __name__ == "__main__":
    """Run the parser for live testing."""
    lexer = FStringLexer()
    parser = FStringParser()
    while True:
        try:
            text = input("fstring > ")
        except EOFError:
            break
        if text:
            tokens = lexer.tokenize(text)
            print("Tokens:")
            for i in tokens:
                print(i, end=", ")
            tokens = lexer.tokenize(text)
            result = parser.parse(tokens)
            print("\nParse Result:")
            print(result)



"""Jaclang setup file."""

from setuptools import find_packages, setup  # type: ignore

VERSION = "0.0.5"

setup(
    name="jaclang",
    version=VERSION,
    packages=find_packages(include=["jaclang", "jaclang.*"]),
    install_requires=[],
    package_data={
        "": ["*.ini", "*.jac"],
    },
    entry_points={
        "console_scripts": [
            "jac = jaclang.cli:cli.start_cli",
        ]
    },
    author="Jason Mars",
    author_email="jason@jaseci.org",
    url="https://github.com/Jaseci-Labs/jaclang",
)



"""Static Import Pass."""
from os import path

import jaclang.jac.absyntree as ast
from jaclang.jac.passes import Pass
from jaclang.jac.passes.blue import SubNodeTabPass


class ImportPass(Pass):
    """Jac statically imports all modules."""

    def before_pass(self) -> None:
        """Run once before pass."""
        self.import_table = {}

    def enter_module(self, node: ast.Module) -> None:
        """Run Importer."""
        self.cur_node = node
        self.terminate()  # Turns off auto traversal for deliberate traversal
        self.run_again = True
        while self.run_again:
            self.run_again = False
            for i in self.get_all_sub_nodes(node, ast.Import):
                if i.lang.value == "jac" and not i.sub_module:
                    self.run_again = True
                    ast.append_node(i, self.import_module(i, node.mod_path))
                    i.sub_module = i.kid[-1]
                self.enter_import(i)
            SubNodeTabPass(mod_path=node.mod_path, input_ir=node)

    def enter_import(self, node: ast.Import) -> None:
        """Sub objects.

        lang: Name,
        path: ModulePath,
        alias: Optional[Name],
        items: Optional[ModuleItems], # Items matched during def/decl pass
        is_absorb: bool,
        self.sub_module = None
        """
        self.cur_node = node
        if node.alias and node.sub_module:
            node.sub_module.name = node.alias.value
        # Items matched during def/decl pass

    # Utility functions
    # -----------------

    def import_module(self, node: ast.Import, mod_path: str) -> ast.AstNode:
        """Import a module."""
        from jaclang.jac.transpiler import jac_file_to_pass
        from jaclang.jac.passes.blue.ast_build_pass import AstBuildPass

        base_dir = path.dirname(mod_path)
        target = path.normpath(
            path.normpath(
                path.join(base_dir, *(node.path.path_str.split("."))) + ".jac"
            )
        )
        if target in self.import_table:
            # self.warning(f"Circular import detected, module {target} already imported.")
            return self.import_table[target]

        if not path.exists(target):
            self.error(f"Could not find module {target}")
        mod_pass = jac_file_to_pass(
            file_path=target, base_dir=base_dir, target=AstBuildPass
        )
        mod = mod_pass.ir
        if isinstance(mod, ast.Module):
            self.import_table[target] = mod
            mod.is_imported = True
        else:
            self.error(f"Module {target} is not a valid Jac module.")
            raise mod_pass.gen_exception(f"Module {target} is not a valid Jac module.")
        return mod
